#!/usr/bin/env node

import chalk from 'chalk';
import chalkAnimation from 'chalk-animation';
import figlet from 'figlet';
import gradient, { pastel } from 'gradient-string';
import inquirer from 'inquirer';
import { execSync } from 'child_process';
import { promises as fs } from 'fs';
import path from 'path';
import { fileURLToPath } from 'url';
import cliProgress from 'cli-progress';

const __filename = fileURLToPath(import.meta.url);
const __dirname = path.dirname(__filename);

// Display a styled title
const showTitle = () => {
  console.log(pastel(figlet.textSync('Discord Bot With NestJS CLI!', { horizontalLayout: 'default' })));
};

// Display a welcome message animation
const showWelcomeMessage = () => {
  const animation = chalkAnimation.rainbow('Welcome to Discord Bot With NestJS CLI! Let’s start 🚀\n');
  return new Promise(resolve => {
    setTimeout(() => {
      animation.stop();
      resolve();
    }, 3000);
  });
};

// Ask user for project details
const askProjectDetails = async () => {
  console.log(chalk.blue('Asking for project details...'));
  const questions = [
    {
      type: 'input',
      name: 'projectName',
      message: 'What is the name of your project?',
      validate: input => input ? true : 'The project name cannot be empty.'
    },
    {
      type: 'list',
      name: 'packageManager',
      message: 'Which package manager do you want to use?',
      choices: ['npm', 'yarn', 'pnpm'],
    }
  ];

  const answers = await inquirer.prompt(questions);
  console.log(chalk.green(`Project name: ${answers.projectName}, Package manager: ${answers.packageManager}`));
  return answers;
};

// Create the NestJS project
const createNestProject = async ({ projectName, packageManager }) => {
  console.log(chalk.blue('Starting project creation...'));
  const progressBar = new cliProgress.SingleBar({
    format: 'Progress |{bar}| {percentage}% ',
    barCompleteChar: '\u2588',
    barIncompleteChar: ' ',
    hideCursor: true
  }, cliProgress.Presets.shades_classic);

  progressBar.start(100, 0);

  try {
    const startTime = Date.now();

    console.log(chalk.blue('Running NestJS CLI to create the project...'));
    progressBar.update(10);
    execSync(`npx @nestjs/cli new ${projectName} -p ${packageManager}`, { stdio: 'ignore' });

    console.log(chalk.blue('Updating package.json description...'));
    progressBar.update(50);

    const packageJsonPath = path.join(projectName, 'package.json');
    const packageJson = JSON.parse(await fs.readFile(packageJsonPath, 'utf-8'));
    packageJson.description = 'Created with Discord Bot With NestJS CLI by Mquiuqui';
    await fs.writeFile(packageJsonPath, JSON.stringify(packageJson, null, 2));

    console.log(chalk.blue('Adding Discord module and service...'));
    progressBar.update(70);

    const discordBotModule = `import { Module } from '@nestjs/common';
import { DiscordService } from './discord.service';

@Module({
  providers: [DiscordService],
})
export class DiscordModule {}`;

    const discordService = `import { Injectable, OnModuleInit } from '@nestjs/common';
import { Client, GatewayIntentBits } from 'discord.js';

@Injectable()
export class DiscordService implements OnModuleInit {
  private readonly client: Client;

  constructor() {
    this.client = new Client({
      intents: [
        GatewayIntentBits.Guilds,
        GatewayIntentBits.GuildVoiceStates,
        GatewayIntentBits.GuildMessages,
        GatewayIntentBits.GuildMessageReactions,
        GatewayIntentBits.MessageContent,
      ],
    });
  }

  async onModuleInit() {
    console.log('Starting Discord bot...');
    await this.client.login(process.env.DISCORD_ENV);

    this.client.on('ready', () => {
      console.log('Discord bot is online!');
    });

    this.client.on('messageCreate', async (message) => {
      if (!message.content.startsWith('!') || message.author.bot) return;

      const args = message.content.slice(1).trim().split(/ +/);
      const command = args.shift()?.toLowerCase();

      if (command === 'test') {
        await message.reply('Test command executed!');
      }
    });
  }
}`;

    const envFileContent = `DISCORD_ENV=your_discord_bot_token_here\n`;

    const discordDirPath = path.join(projectName, 'src/discord');
    const modulePath = path.join(discordDirPath, 'discord.module.ts');
    const servicePath = path.join(discordDirPath, 'discord.service.ts');
    const envPath = path.join(projectName, '.env');

    console.log(chalk.blue('Creating necessary files and directories...'));
    await fs.mkdir(discordDirPath, { recursive: true });
    await fs.writeFile(modulePath, discordBotModule);
    await fs.writeFile(servicePath, discordService);
    await fs.writeFile(envPath, envFileContent);

    progressBar.update(90);

    console.log(chalk.blue('Installing discord.js package...'));
    execSync(`cd ${projectName} && ${packageManager} add discord.js`, { stdio: 'ignore' });

    progressBar.update(100);
    progressBar.stop();

    const elapsedTime = ((Date.now() - startTime) / 1000).toFixed(2);
    console.log(chalk.green(`Project ${projectName} successfully created in ${elapsedTime}s!`));
  } catch (error) {
    progressBar.stop();
    console.error(chalk.red('Failed to create the NestJS project.'), error);
    process.exit(1);
  }
};

// Commands
const showVersion = async () => {
    const packageJson = await import('./package.json', { assert: { type: 'json' } });
    console.log(chalk.green(`Discord Bot With NestJS CLI version: ${packageJson.default.version}`));
  };
  const showHelp = () => {
    console.log(`
  ${chalk.blue('Available commands:')}`);
    console.log(`${chalk.cyan('init')}     - Create a new Discord Bot project`);
    console.log(`${chalk.cyan('version')}  - Display the CLI version`);
    console.log(`${chalk.cyan('help')}     - Show this help message`);
  };
  
  const main = async () => {
    const args = process.argv.slice(2);
  
    if (args.length === 0 || args.includes('help')) {
      showHelp();
      return;
    }
  
    if (args.includes('version')) {
      return showVersion();
    }
  
    if (args.includes('init')) {
      console.clear();
      console.log(chalk.blue('Launching Discord Bot With NestJS CLI...'));
      showTitle();
      await showWelcomeMessage();
  
      console.log(chalk.blue('Gathering project details...'));
      const projectDetails = await askProjectDetails();
  
      console.log(chalk.blue('Starting project creation process...'));
      await createNestProject(projectDetails);
  
      console.log(pastel(`\n✨ Project ${projectDetails.projectName} is ready to use!`));
      console.log(chalk.blue('\nRun the following commands to get started:'));
      console.log(chalk.cyan(`  cd ${projectDetails.projectName}`));
      console.log(chalk.cyan(`  ${projectDetails.packageManager === 'npm' ? 'npm run' : projectDetails.packageManager} start:dev`));
      return;
    }
  
    console.log(chalk.red(`Unknown command: ${args[0]}`));
    showHelp();
  };
  
  main().catch(error => {
    console.error(chalk.red('Error running the CLI:'), error);
    process.exit(1);
  });
  